-- https://github.com/moody/DethsDBLib

local LibName, LibVersion = "DethsDBLib", "1.1"
local PREFIX = format("%s_%s_", LibName, LibVersion)

-- DethsLibLoader
local DDBL = DethsLibLoader:Create(LibName, LibVersion)
if not DDBL then return end

-- Upvalues
local _G, assert, format, type, tsort =
      _G, assert, format, type, table.sort

local getmetatable, next, pairs, setmetatable =
      getmetatable, next, pairs, setmetatable

local GetAddOnMetadata, IsAddOnLoaded = GetAddOnMetadata, IsAddOnLoaded

-- Consts
local PLAYER_KEY = format("%s-%s", UnitName("PLAYER"), GetRealmName())
local TOC_ENTRY = "X-DethsDBLib"
local TOC_ERROR_MSG = [=[
%s.toc must contain:

## SavedVariables: my_addon_sv_key
## X-DethsDBLib: my_addon_sv_key

Where "my_addon_sv_key" is a unique global variable name.
]=]

-- Variables
local Addon_Cache = {} -- list of addon names previously used to create a db
local DB_Cache = {} -- array of database tables
local DB_Mixins = {} -- database table mixins

-- ============================================================================
-- Local Helper Functions
-- ============================================================================

-- Returns a copy of the specified table.
-- @param t - the table to copy
-- @param deep - if true, the table will be copied recursively
local function tcopy(t, deep)
  if (t == nil) then return nil end
  local copy = {}
  -- Copy
  for k, v in pairs(t) do
    if deep and (type(v) == "table") then
      copy[k] = tcopy(v, deep)
    else
      copy[k] = v
    end
  end
  -- Copy metatable
  setmetatable(copy, tcopy(getmetatable(t), deep))
  -- Return
  return copy
end

-- Populates a table with missing default key-value pairs.
-- @param t - the table to add to
-- @param defaults - the table with default values
local function populate_t(t, defaults)
  for k, v in pairs(defaults) do
    if (type(v) == "table") then
      if (type(t[k]) == "table") then
        populate_t(t[k], v) -- recursively populate
      else
        t[k] = tcopy(v, true) -- deep copy
      end
    elseif (t[k] == nil) then
      t[k] = v  -- add default value
    end
  end
end

-- Depopulates a table by removing default key-value pairs.
-- @param t - the table to remove from
-- @param defaults - the table with default values
local function depopulate_t(t, defaults)
  for k, v in pairs(t) do
    if (type(v) == "table") and (type(defaults[k]) == "table") then
       -- Recursively depopulate
      depopulate_t(v, defaults[k])
      -- If table is now empty, remove it
      if (next(v) == nil) then t[k] = nil end
    elseif (defaults[k] == v) then
      t[k] = nil -- remove default value
    end
  end
end

-- Updates the specified SavedVariables table for the current version of DDBL.
-- @param svar - the saved variables table to update
local function update_svar_version(svar)
  if (svar.DDBL_Version == LibVersion) then return end
  local version = tonumber(svar.DDBL_Version or "")

  ----------------
  -- 1.0 -> 1.1 --
  ----------------
  if not svar.DDBL_Version or (version <= 1.0) then
    -- Move _current_profile_key profile entries to svar.ProfileKeys
    for key, profile in pairs(svar.Profiles) do
      local profileKey = profile._current_profile_key
      if profileKey then
        svar.ProfileKeys[key] = profileKey
        profile._current_profile_key = nil
      end
    end
  end

  svar.DDBL_Version = LibVersion
end

-- Initializes the database's underlying SavedVariables table.
-- @param db - the database to initialize
local function initialize_db(db)
  local svar, defaults = db._svar, db._defaults

  -- Create the ProfileKeys table if it doesn't exist
  if (type(svar.ProfileKeys) ~= "table") then svar.ProfileKeys = {} end

  -- Create or populate the Global table
  if (type(svar.Global) ~= "table") then
    svar.Global = tcopy(defaults.Global, true)
  else -- Populate
    populate_t(svar.Global, defaults.Global)
  end
  db.Global = svar.Global

  -- Create the Profiles table if it doesn't exist
  if (type(svar.Profiles) ~= "table") then
    svar.Profiles = {}
  else -- Populate each profile
    for key, profile in pairs(svar.Profiles) do
      populate_t(profile, defaults.Profile)
    end
  end
  
  update_svar_version(svar)

  -- Initialize db.Profile
  local profileKey = svar.ProfileKeys[PLAYER_KEY] or PLAYER_KEY
  local profile = svar.Profiles[profileKey]
  db.Profile = profile
  db._profile_key = profileKey

  if not profile then -- Revert to player profile
    svar.ProfileKeys[PLAYER_KEY] = nil
    db._profile_key = PLAYER_KEY
    
    if svar.Profiles[PLAYER_KEY] then
      db.Profile = svar.Profiles[PLAYER_KEY]
    else -- Create new profile for player
      profile = tcopy(defaults.Profile, true)
      svar.Profiles[PLAYER_KEY] = profile
      db.Profile = profile
    end
  end
end

-- ============================================================================
-- DB_Mixins: SavedVariable Getters & Setters
-- ============================================================================

do
  local keys = {
    -- [1] = a nested table
    -- ...
    -- [n] = key in nested table
  }

  -- Returns the value, table, and key for a specified key in a specified table.
  -- @param t - the table
  -- @param k - the key, can be a nested value (e.g. "A.B.C")
  -- @return 1 - the key value
  -- @return 2 - the table the key resides in
  -- @return 3 - the key
  local function get(t, k)
    if k:find(".", 1, true) then
      for key in pairs(keys) do keys[key] = nil end
      -- Split keys by dot operator (e.g. "A.B.C" = {"A", "B", "C"})
      for key in k:gmatch("([^.]+)") do keys[#keys+1] = key end
      
      -- If key is nested, update ref to nested table and key
      if (#keys > 1) then
        for i=1, (#keys - 1) do t = t[keys[i]] end
        k = keys[#keys]
      end
    end

    return t[k], t, k
  end

  -- Sets the specified key in the global or per char SVs to the specified value.
  -- @param t - the table
  -- @param k - the key, can be a nested value (e.g. "A.B.C")
  -- @param v - the new value
  local function set(t, k, v)
    local _, nt, nk = get(t, k)
    nt[nk] = v
  end

  -- Returns the value of the specified key in the global SVs.
  -- @param k - the key, can be a nested value (e.g. "A.B.C")
  function DB_Mixins:GetGlobal(k)
    local v = get(self.Global, k)
    return v
  end

  -- Sets the specified key in the global SVs to the specified value.
  -- @param k - the key, can be a nested value (e.g. "A.B.C")
  -- @param v - the new value
  function DB_Mixins:SetGlobal(k, v)
    set(self.Global, k, v)
  end

  -- Returns the value of the specified key in the current profile SVs.
  -- @param k - the key, can be a nested value (e.g. "A.B.C")
  function DB_Mixins:Get(k)
    local v = get(self.Profile, k)
    return v
  end

  -- Sets the specified key in the current profile SVs to the specified value.
  -- @param k - the key, can be a nested value (e.g. "A.B.C")
  -- @param v - the new value
  function DB_Mixins:Set(k, v)
    set(self.Profile, k, v)
  end
end

-- ============================================================================
-- DB_Mixins: General Functions
-- ============================================================================

-- Creates a new profile with the specified key.
-- @param key - the key for the profile to be created
-- @param importTable - a table of existing profile settings to use [optional]
-- @return - true if the operation is successful
function DB_Mixins:CreateProfile(key, importTable)
  if not key or self._svar.Profiles[key] then return false end
  if (type(importTable) == "table") then
    importTable = tcopy(importTable, true)
    populate_t(importTable, self._defaults.Profile)
    self._svar.Profiles[key] = importTable
  else
    self._svar.Profiles[key] = tcopy(self._defaults.Profile, true)
  end
  return true
end

-- Copies a profile by key, overwriting the current profile's settings.
-- @param key - the key of the profile to copy
-- @return - true if the operation is successful
function DB_Mixins:CopyProfile(key)
  if not key or (key == self._profile_key) then return false end
  local profile = self._svar.Profiles[key]
  if profile then
    -- Deep copy profile
    local copy = tcopy(profile, true)
    -- Set profile key to reference new table
    self._svar.Profiles[self._profile_key] = copy
    -- Set current profile
    self.Profile = copy
    return true
  end
  return false
end

-- Deletes a profile by key.
-- @param key - the key of the profile to delete
-- @return - true if the operation is successful
function DB_Mixins:DeleteProfile(key)
  if not key or (key == self._profile_key) then return false end
  self._svar.Profiles[key] = nil
  return true
end

-- Sets the current profile by key, creating it if necessary. Reverts to the
-- player profile if the key is not specified.
-- @param key - the key of the profile to be set [optional]
-- @return - true if the operation is successful
function DB_Mixins:SetProfile(key)
  if not key then key = PLAYER_KEY end
  if (key == self._profile_key) then return false end
  -- Set profile
  self:CreateProfile(key) -- Does nothing if profile already exists
  self.Profile = self._svar.Profiles[key]
  self._profile_key = key
  -- Set current profile key in svar.ProfileKeys
  self._svar.ProfileKeys[PLAYER_KEY] = (key ~= PLAYER_KEY) and key or nil
  return true
end

-- Returns the key of the player profile.
function DB_Mixins:GetPlayerKey()
  return PLAYER_KEY
end

-- Returns the key of the current profile.
function DB_Mixins:GetProfileKey()
  return self._profile_key
end

-- Returns a sorted table of profile keys.
function DB_Mixins:GetProfileKeys()
  local t = {}
  for k in pairs(self._svar.Profiles) do t[#t+1] = k end
  tsort(t)
  return t
end

-- Returns true if a profile with the specified key exists.
-- @param key - the key of the profile to check for
function DB_Mixins:ProfileExists(key)
  return (type(self._svar.Profiles[key]) == "table")
end

-- ============================================================================
-- DB Creation Function
-- ============================================================================

-- Creates and returns a new database table.
-- @param addonName - the name of the addon to create a database for
-- @param defaults - a table of default key-value pairs for the database
function DDBL:Create(addonName, defaults)
  local svarKey -- set during validation

  do -- Validation
    -- Validate addonName string
    if not (type(addonName) == "string") then
      error("addonName must be a string")
    end
    -- Validate that the addon exists
    if not IsAddOnLoaded(addonName) then
      error(format("addon not found: \"%s\"", addonName))
    end
    -- Validate that a database for the addon has not been created
    if Addon_Cache[addonName] then
      error(format("database for \"%s\" already exists", addonName))
    end
    -- Validate defaults table
    if not (type(defaults) == "table") then
      error("defaults must be a table")
    elseif not (type(defaults.Global) == "table") then
      error("defaults.Global must be a table")
    elseif not (type(defaults.Profile) == "table") then
      error("defaults.Profile must be a table")
    end
    -- Validate .toc has entry for DDBL
    svarKey = GetAddOnMetadata(addonName, TOC_ENTRY)
    if not (type(svarKey) == "string") then
      error(format(TOC_ERROR_MSG, addonName))
    end
  end

  -- Add addonName to cache
  Addon_Cache[addonName] = true

  -- Get SavedVariables table, create it if necessary
  local svar = _G[svarKey]
  if not (type(svar) == "table") then
    local t = {}
    _G[svarKey] = t
    svar = t
  end
  
  -- Create DB table
  local db = {
    -- DB_Mixin functions
    -- Global = ref to svar.Global table
    -- Profile = ref to table in svar.Profiles
    -- _profile_key = key of Profile table in svar.Profiles
    _svar = svar,
    _defaults = defaults
  }
  -- Add mixins
  for k, v in pairs(DB_Mixins) do db[k] = v end
  -- Initialize
  initialize_db(db)
  -- Store database
  DB_Cache[#DB_Cache+1] = db
  -- Return
  return db
end

-- ============================================================================
-- Frame
-- ============================================================================

local frame = CreateFrame("Frame", PREFIX.."EventFrame")

function frame:OnEvent(event, ...)
  if (event == "PLAYER_LOGOUT") then
    -- Remove all default key-value pairs from databases
    for _, db in pairs(DB_Cache) do
      local svar, defaults = db._svar, db._defaults

      -- Wipe defaults from global
      depopulate_t(svar.Global, defaults.Global)
      if not next(svar.Global) then svar.Global = nil end

      -- Wipe defaults from profiles
      for k, profile in pairs(svar.Profiles) do
        depopulate_t(profile, defaults.Profile)
      end
      if not next(svar.Profiles) then svar.Profiles = nil end
    end
  end
end

frame:SetScript("OnEvent", frame.OnEvent)
frame:RegisterEvent("PLAYER_LOGOUT")
