local AddonName, Addon = ...

-- Lib
local DGL = Addon.Libs.DGL

-- Upvalues
local pairs, min, max, random = pairs, math.min, math.max, math.random

local Clamp = Clamp

-- Game
local Column = Addon.Column
Column.Speed = 350

local Game = Addon.Game
local Spritesheet = Addon.Spritesheet
local Sublayers = Addon.Sublayers
local World = Addon.World

-- Mixins
local Mixins = {}

-- Consts
local SPAWN_DELAY_MIN, SPAWN_DELAY_MAX = 1, 1.75
local TEX_COORD = Spritesheet.Sprites.Column.Coords
local WIDTH = Spritesheet.Sprites.Column.Width
local SCORE_THRESHOLD = Game.HALF_WIDTH - WIDTH - (Spritesheet.Sprites.BirdUp.Width * 0.5)
local TYPES = {
  "DOWN",
  "UP",
  "BOTH"
}

-- Variables
local columns = {}
local timer, delay = 0, 0

-- ============================================================================
-- Local Functions
-- ============================================================================

-- Creates a column.
-- @number y - the y coordinate for the column
-- @number height - the height of the column
-- @number rotation - the rotation of the column
-- @boolean scored - if true the column will not increase the score when cleared
local function createColumn(y, height, rotation, scored)
  local column = DGL.GameObject:Create(Game, "Column", Game.WIDTH + WIDTH, y, nil, Sublayers.Column)
  column:SetWidth(WIDTH)
  column:SetHeight(height)
  column.Texture:SetTexture(Spritesheet.Path)
  column.Texture:SetTexCoord(TEX_COORD[1], TEX_COORD[2], TEX_COORD[3], min(height / Spritesheet.Height, TEX_COORD[4]))
  column.Texture:SetRotation(rotation)
  column.scored = scored
  -- Add collision rect to world
  World:add(column, column.X, column.Y, WIDTH, height)
  -- Add mixins
  for k, v in next, Mixins do column[k] = v end
  -- Store and return
  columns[column] = true
end

-- Returns a random delay time for spawning columns.
local function getRandomDelay()
  return min(max(1 + random(), SPAWN_DELAY_MIN) + random(), SPAWN_DELAY_MAX)
end

-- ============================================================================
-- General Functions
-- ============================================================================

-- Resets spawn timers and disposes all columns.
function Column:Reset()
  timer, delay = 0, 0
  -- Dispose all columns
  for column in pairs(columns) do
    column:Dispose()
    columns[column] = nil
  end
end

-- Spawns a column with random properties.
function Column:Spawn()
  local type = TYPES[random(#TYPES)]
  if (type == "BOTH") then
    -- Top column
    local top_height = random(Game.HEIGHT * 0.15, Game.HEIGHT * 0.45)
    createColumn(0, top_height, PI)
    -- Bottom column
    local bottom_height = Game.HEIGHT - top_height - (Game.HEIGHT * 0.2)
    createColumn(Game.HEIGHT - bottom_height, bottom_height, 0, true)
  else -- DOWN or UP
    local height = random(Game.HEIGHT * 0.375, Game.HEIGHT * 0.5) -- Min = 37.5%, Max = 50%
    local y = (type == "DOWN") and 0 or Game.HEIGHT - (Spritesheet.Sprites.Ground.Height - 10) - height
    local rotation = (type == "DOWN") and PI or 0
    createColumn(y, height, rotation)
  end
end

-- Updates the spawning of columns.
function Column:OnUpdate(elapsed)
  timer = timer + elapsed
  if (timer >= delay) then
    timer = 0
    delay = getRandomDelay()
    self:Spawn()
  end
end

-- ============================================================================
-- Mixins
-- ============================================================================

function Mixins:OnUpdate(elapsed)
  if (Game:GetState() ~= Game.States.Playing) then return end

  -- Move left
  local cols, cols_len
  self.X, self.Y, cols, cols_len = World:move(self, self.X - (Column.Speed * elapsed), self.Y)
  for i=1, cols_len do
    if (cols[i].other.Name == "Player") then Game:GameOver() end
  end

  -- Check for score
  if not self.scored and (self.X < SCORE_THRESHOLD) then
    self.scored = true
    Game:Score()
  end
  
  -- Dispose when no longer visible
  if (self.X < -self.WIDTH) then
    self:Dispose()
    columns[self] = nil
  end
end

function Mixins:OnDispose()
  World:remove(self)
end
